{#- Default system message if no system prompt is passed. #}
{%- set default_system_message = '' %}

{#- Begin of sequence token. #}
{{- bos_token }}


{#- Handle system prompt if it exists. #}
{#- Extract all system messages and merge into system prompt. #}
{%- set ns_sys = namespace(system_parts=[], filtered=[]) %}
{%- for message in messages %}
    {%- if message['role'] == 'system' %}
        {%- if message['content'] is string %}
            {%- set ns_sys.system_parts = ns_sys.system_parts + [message['content']] %}
        {%- else %}
            {%- for block in message['content'] %}
                {%- if block['type'] == 'text' %}
                    {%- set ns_sys.system_parts = ns_sys.system_parts + [block['text']] %}
                {%- else %}
                    {{- raise_exception('Only text chunks are supported in system message content.') }}
                {%- endif %}
            {%- endfor %}
        {%- endif %}
    {%- else %}
        {%- set ns_sys.filtered = ns_sys.filtered + [message] %}
    {%- endif %}
{%- endfor %}
{%- if ns_sys.system_parts | length > 0 %}
    {%- set system_message = ns_sys.system_parts | join('\n\n') %}
{%- else %}
    {%- set system_message = default_system_message %}
{%- endif %}
{%- set loop_messages = ns_sys.filtered %}


{#- Tools definition #}
{%- set available_tools = '' %}
{%- set has_tools = false %}
{%- if tools is defined and tools is not none and tools|length > 0 %}
    {%- set has_tools = true %}
    {%- set available_tools = '[AVAILABLE_TOOLS]' + (tools| tojson) + '[/AVAILABLE_TOOLS]' %}
{%- endif %}

{#- Macros #}
{%- macro render_content(content, context_name, supported_types_desc, support_images) -%}
    {%- if content is string -%}
        {{- content -}}
    {%- elif content -%}
        {%- for block in content -%}
            {%- if block['type'] == 'text' -%}
                {{- block['text'] -}}
            {%- elif support_images and block['type'] in ['image', 'image_url'] -%}
                {{- '[IMG]' -}}
            {%- else -%}
                {{- raise_exception('Only ' + supported_types_desc + ' chunks are supported in ' + context_name + '.') -}}
            {%- endif -%}
        {%- endfor -%}
    {%- else -%}
        {{- raise_exception(context_name + ' must have non-empty content.') -}}
    {%- endif -%}
{%- endmacro -%}

{#- Aggregate consecutive messages with the same role except system and tool. #}
{#- A sentinel message is appended so the last group gets flushed inside the loop. #}
{%- set ns_agg = namespace(messages=[], current_group=[], current_role=none) %}
{%- for message in loop_messages + [{'role': '__sentinel__'}] %}
    {%- if message['role'] != ns_agg.current_role or message['role'] == 'system' or message['role'] == 'tool' %}
        {%- if ns_agg.current_role == 'tool' %}
            {%- set ns_agg.messages = ns_agg.messages + ns_agg.current_group %}
        {%- elif ns_agg.current_role is not none %}
            {%- set ns_c = namespace(text_parts=[], chunks=[], has_non_text=false, tool_calls=[]) %}
            {%- for msg in ns_agg.current_group %}
                {%- if msg['content'] is string %}
                    {%- set ns_c.text_parts = ns_c.text_parts + [msg['content']] %}
                {%- elif msg['content'] is not none %}
                    {%- for block in msg['content'] %}
                        {%- if block['type'] == 'text' %}
                            {%- set ns_c.text_parts = ns_c.text_parts + [block['text']] %}
                        {%- else %}
                            {%- if ns_c.text_parts | length > 0 %}
                                {%- set ns_c.chunks = ns_c.chunks + [{'type': 'text', 'text': ns_c.text_parts | join('\n\n')}] %}
                                {%- set ns_c.text_parts = [] %}
                            {%- endif %}
                            {%- set ns_c.chunks = ns_c.chunks + [block] %}
                            {%- set ns_c.has_non_text = true %}
                        {%- endif %}
                    {%- endfor %}
                {%- endif %}
                {%- if msg['tool_calls'] is defined and msg['tool_calls'] is not none %}
                    {%- set ns_c.tool_calls = ns_c.tool_calls + msg['tool_calls'] | list %}
                {%- endif %}
            {%- endfor %}
            {%- if ns_c.has_non_text %}
                {%- if ns_c.text_parts | length > 0 %}
                    {%- set ns_c.chunks = ns_c.chunks + [{'type': 'text', 'text': ns_c.text_parts | join('\n\n')}] %}
                {%- endif %}
                {%- set merged_content = ns_c.chunks %}
            {%- else %}
                {%- set merged_content = ns_c.text_parts | join('\n\n') %}
            {%- endif %}
            {%- if ns_c.tool_calls | length > 0 %}
                {%- set ns_agg.messages = ns_agg.messages + [{'role': ns_agg.current_role, 'content': merged_content, 'tool_calls': ns_c.tool_calls}] %}
            {%- else %}
                {%- set ns_agg.messages = ns_agg.messages + [{'role': ns_agg.current_role, 'content': merged_content}] %}
            {%- endif %}
        {%- endif %}
        {%- if message['role'] != '__sentinel__' %}
            {%- set ns_agg.current_group = [message] %}
            {%- set ns_agg.current_role = message['role'] %}
        {%- endif %}
    {%- else %}
        {%- set ns_agg.current_group = ns_agg.current_group + [message] %}
    {%- endif %}
{%- endfor %}
{%- set loop_messages = ns_agg.messages %}

{#- Validates message ordering. #}
{%- set ns = namespace(max_idx_user=-1) %}
{%- if loop_messages | length > 0 and loop_messages[0]['role'] != 'user' %}
    {{- raise_exception('Conversation must start with a user message, got ' + loop_messages[0]['role'] + '.') }}
{%- endif %}
{%- set ns_order = namespace(previous_role=none) %}
{%- for message in loop_messages %}
    {%- set current_role = message['role'] %}
    {%- if ns_order.previous_role is not none %}
        {%- if ns_order.previous_role == 'user' %}
            {%- if current_role not in ['assistant', 'user'] %}
                {{- raise_exception('Unexpected role \'' + current_role + '\' after role \'' + ns_order.previous_role + '\'') }}
            {%- endif %}
        {%- elif ns_order.previous_role == 'assistant' %}
            {%- if current_role not in ['assistant', 'user', 'tool'] %}
                {{- raise_exception('Unexpected role \'' + current_role + '\' after role \'' + ns_order.previous_role + '\'') }}
            {%- endif %}
        {%- elif ns_order.previous_role == 'tool' %}
            {%- if current_role not in ['assistant', 'tool', 'user'] %}
                {{- raise_exception('Unexpected role \'' + current_role + '\' after role \'' + ns_order.previous_role + '\'') }}
            {%- endif %}
        {%- endif %}
    {%- endif %}
    {%- if message.role == 'user' %}
        {%- set ns.max_idx_user = ns.max_idx_user + 1 %}
    {%- endif %}
    {%- set ns_order.previous_role = current_role %}
{%- endfor %}


{%- set ns.index = 0 %}
{#- Handle conversation messages. #}
{%- for message in loop_messages %}
    {#- User messages supports text, image and image_url content. #}
    {%- if message['role'] == 'user' %}
        {%- if (ns.index == ns.max_idx_user) and has_tools %}
            {{- available_tools }}
        {%- endif %}
        {{- '[INST]' }}
        {%- if (ns.index == ns.max_idx_user) and system_message != '' %}
            {{- system_message + '\n\n' }}
        {%- endif %}
        {%- if message['content'] is string %}
            {{- render_content(content=message['content'], context_name='user message content', supported_types_desc='text, image and image_url', support_images=true) -}}
        {%- elif message['content'] | length > 0 %}
            {#- When content has exactly one image and one text block, put image first. #}
            {%- if message['content'] | length == 2 and message['content'][0]['type'] == 'text' and message['content'][1]['type'] in ['image', 'image_url'] %}
                {%- set blocks = [message['content'][1], message['content'][0]] %}
            {%- else %}
                {%- set blocks = message['content'] %}
            {%- endif %}
            {{- render_content(content=blocks, context_name='user message content', supported_types_desc='text, image and image_url', support_images=true) -}}
        {%- else %}
            {{- raise_exception('User message must have a string or a list of chunks in content') }}
        {%- endif %}
        {{- '[/INST]' }}

    {#- Assistant messages supports text content. #}
    {%- elif message['role'] == 'assistant' %}
        {%- if message['content'] is not none and message['content'] | length > 0 and message['tool_calls'] is defined and message['tool_calls'] is not none and message['tool_calls']|length > 0 %}
            {{- raise_exception('Assistant message cannot have both content and tool calls.') }}
        {%- endif %}

        {%- if (message['content'] is none or message['content'] == '' or message['content']|length == 0) and (message['tool_calls'] is not defined or message['tool_calls'] is none or message['tool_calls']|length == 0) %}
            {{- raise_exception('Assistant message must have a string or a list of chunks in content or a list of tool calls.') }}
        {%- endif %}

        {%- if message['content'] %}
            {{- render_content(content=message['content'], context_name='assistant message contents', supported_types_desc='text', support_images=false) -}}
        {%- endif %}

        {%- if message['tool_calls'] is defined and message['tool_calls'] is not none and message['tool_calls']|length > 0 %}
            {{- '[TOOL_CALLS][' }}
            {%- for tool in message['tool_calls'] %}
                {%- set name = tool['function']['name'] %}
                {%- set arguments = tool['function']['arguments'] %}
                {%- set id = tool['id']%}
                {%- if id is not defined or id|length != 9 %}
                    {{- raise_exception('Tool call must have an id of 9 characters or numbers.') }}
                {%- endif %}
                {%- if arguments is not string %}
                    {%- set arguments = arguments|tojson|safe %}
                {%- elif arguments == '' %}
                    {%- set arguments = '{}' %}
                {%- endif %}
                {{- '{"name": "' + name + '", "arguments": ' + arguments + ', "id": "' + id + '"}' }}
                {%- if loop.length > 1 and loop.index < loop.length %}
                    {{- ', ' }}
                {%- endif %}
            {%- endfor %}
            {{- ']' }}
        {%- endif %}

        {{- eos_token }}

    {#- Tool messages only supports text content. #}
    {%- elif message['role'] == 'tool' %}
        {# Try to parse 'content' as int or float if possible #}
        {%- set tool_content = message['content']|string %}
        {# Try to parse as int #}
        {%- set parsed_int = message['content']|int %}
        {% if parsed_int|string == message['content'] %}
            {%- set tool_content = parsed_int %}
        {# If int fails, try to parse as float #}
        {%- else %}
            {%- set parsed_float = message['content']|float %}
            {%- if parsed_float|string == message['content'] %}
                {%- set tool_content = parsed_float %}
            {%- endif %}
        {%- endif %}
        
        {%- if message['call_id'] is not undefined and message['call_id']|length == 9 %}
            {%- set tool_id = message['call_id'] %}
        {%- elif message['tool_call_id'] is not undefined and message['tool_call_id']|length == 9  %}
            {%- set tool_id = message['tool_call_id'] %}
        {%- else %}
            {{- raise_exception('Tool message must have a call_id or tool_call_id of 9 characters or numbers.') }}
        {%- endif %}

        {%- set tool_message = {'content': tool_content, 'call_id': tool_id} %}
        
        {{- '[TOOL_RESULTS]' + (tool_message|tojson) + '[/TOOL_RESULTS]' }}


    {#- Raise exception for unsupported roles. #}
    {%- else %}
        {{- raise_exception('Only user, assistant and tool roles are supported, got ' + message['role'] + '.') }}
    {%- endif %}
    {%- if message['role'] == 'user' %}
        {%- set ns.index = ns.index + 1 %}
    {%- endif %}
{%- endfor %}
